"""Test the publish and consume APIs on a real broker running on localhost."""

from collections import defaultdict
import multiprocessing
import threading
import time
import unittest
import uuid
import socket

from twisted.internet import reactor, task
import mock
import pika
import pkg_resources
import pytest
import pytest_twisted

from fedora_messaging import api, message, exceptions
from fedora_messaging.twisted import service
from fedora_messaging.twisted.protocol import _pika_version


class PubSubTests(unittest.TestCase):
    def test_pub_sub_default_settings(self):
        """
        Assert publishing and subscribing works with the default configuration.

        This should work because the publisher uses the 'amq.topic' exchange by
        default and the consumer also uses the 'amq.topic' exchange with its
        auto-named queue and a default subscription key of '#'.
        """

        # Consumer setup
        def counting_callback(message, storage=defaultdict(int)):
            storage[message.topic] += 1
            if storage[message.topic] == 3:
                raise exceptions.HaltConsumer()

        consumer_process = multiprocessing.Process(
            target=api.consume, args=(counting_callback,)
        )
        msg = message.Message(
            topic=u"nice.message",
            headers={u"niceness": u"very"},
            body={u"encouragement": u"You're doing great!"},
        )

        consumer_process.start()
        # Allow the consumer time to create the queues and bindings
        time.sleep(5)

        for _ in range(0, 3):
            try:
                api.publish(msg)
            except exceptions.ConnectionException:
                consumer_process.terminate()
                self.fail("Failed to publish message, is the broker running?")

        consumer_process.join(timeout=30)
        self.assertEqual(0, consumer_process.exitcode)

    @mock.patch("fedora_messaging.api._session_cache", threading.local())
    def test_pub_connection_refused(self):
        """Assert ConnectionException is raised on connection refused."""
        # Because we don't call accept, we can be sure of a connection refusal
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("", 0))
        url = "amqp://localhost:{port}/".format(port=sock.getsockname()[1])
        api._session_cache.session = api._session.PublisherSession(amqp_url=url)

        self.assertRaises(exceptions.ConnectionException, api.publish, api.Message())


@pytest.mark.skipif(
    _pika_version < pkg_resources.parse_version("1.0.0b1"),
    reason="Twisted with confirms only supported on pika-1.0.0b1+",
)
@pytest_twisted.inlineCallbacks
def test_check_confirms():
    """Assert confirmations are enabled by default."""
    serv = service.FedoraMessagingService(amqp_url="amqp://")
    serv.startService()
    client = yield serv.getFactory().whenConnected()
    channel = yield client._allocate_channel()
    assert channel._delivery_confirmation is True
    serv.stopService()


@pytest_twisted.inlineCallbacks
def test_basic_pub_sub():
    """Basic test of the Twisted publishing/subscribing support"""
    queue = str(uuid.uuid4())
    queues = [
        {"queue": queue, "auto_delete": True, "arguments": {"x-expires": 60 * 1000}}
    ]
    msg = message.Message(
        topic=u"nice.message",
        headers={u"niceness": u"very"},
        body={u"encouragement": u"You're doing great!"},
    )
    expected_headers = {
        u"fedora_messaging_severity": 20,
        u"fedora_messaging_schema": u"base.message",
        u"niceness": u"very",
    }
    messages_received = []
    serv = service.FedoraMessagingService(amqp_url="amqp://")

    serv.startService()
    client = yield serv.getFactory().whenConnected()
    yield client.declare_queues(queues)
    yield client.bind_queues(
        [{"queue": queue, "exchange": "amq.topic", "routing_key": "#"}]
    )

    def callback(message):
        messages_received.append(message)
        if len(messages_received) == 3:
            raise exceptions.HaltConsumer()

    yield client.consume(callback, queue)
    for _ in range(0, 3):
        yield client.publish(msg, "amq.topic")
    yield task.deferLater(reactor, 3.0, lambda: True)
    serv.stopService()

    assert len(messages_received) == 3
    for m in messages_received:
        assert u"nice.message" == m.topic
        assert {u"encouragement": u"You're doing great!"} == m.body
        assert "sent-at" in m._headers
        del m._headers["sent-at"]
        assert expected_headers == m._headers


@pytest_twisted.inlineCallbacks
def test_unhandled_exception_cancels_consumer():
    """Assert any unhandled Exception results in the consumer being canceled."""
    queue = str(uuid.uuid4())
    queues = [
        {"queue": queue, "auto_delete": True, "arguments": {"x-expires": 60 * 1000}}
    ]
    serv = service.FedoraMessagingService(amqp_url="amqp://")

    serv.startService()
    client = yield serv.getFactory().whenConnected()
    yield client.declare_queues(queues)
    yield client.bind_queues(
        [{"queue": queue, "exchange": "amq.topic", "routing_key": "#"}]
    )

    def callback(message):
        raise Exception("Panic!")

    yield client.consume(callback, queue)
    assert len(client._consumers) == 1

    yield client.publish(message.Message(), "amq.topic")
    yield task.deferLater(reactor, 3.0, lambda: True)
    assert len(client._consumers) == 0
    serv.stopService()


@pytest_twisted.inlineCallbacks
def test_nack_handled():
    """Assert raising Nack in a consumer works and messages are re-delivered"""
    queue = str(uuid.uuid4())
    queues = [
        {"queue": queue, "auto_delete": True, "arguments": {"x-expires": 60 * 1000}}
    ]
    messages = []
    serv = service.FedoraMessagingService(amqp_url="amqp://")

    serv.startService()
    client = yield serv.getFactory().whenConnected()
    yield client.declare_queues(queues)
    yield client.bind_queues(
        [{"queue": queue, "exchange": "amq.topic", "routing_key": "#"}]
    )

    def callback(message):
        messages.append(message)
        if len(messages) < 3:
            raise exceptions.Nack()

    yield client.consume(callback, queue)
    assert len(client._consumers) == 1

    yield client.publish(message.Message(), "amq.topic")
    yield task.deferLater(reactor, 3.0, lambda: True)

    assert len(messages) == 3
    assert len(set([m.id for m in messages])) == 1
    assert len(client._consumers) == 1

    yield client.cancel(queue)
    serv.stopService()


@pytest_twisted.inlineCallbacks
def test_drop_handled():
    """Assert raising Drop in a consumer works and messages are not re-delivered"""
    queue = str(uuid.uuid4())
    messages = []
    serv = service.FedoraMessagingService(amqp_url="amqp://")
    serv.startService()
    client = yield serv.getFactory().whenConnected()
    queues = [
        {"queue": queue, "auto_delete": True, "arguments": {"x-expires": 60 * 1000}}
    ]
    yield client.declare_queues(queues)
    yield client.bind_queues(
        [{"queue": queue, "exchange": "amq.topic", "routing_key": "#"}]
    )

    def callback(message):
        messages.append(message)
        raise exceptions.Drop()

    yield client.consume(callback, queue)
    assert len(client._consumers) == 1

    yield client.publish(message.Message(), "amq.topic")
    yield task.deferLater(reactor, 3.0, lambda: True)  # Just wait a few seconds

    assert len(messages) == 1
    assert len(client._consumers) == 1
    yield client.cancel(queue)
    serv.stopService()


@pytest_twisted.inlineCallbacks
def test_declare_queue_failures():
    """Assert that if a queue can't be declared, it results in an exception."""
    serv = service.FedoraMessagingService(amqp_url="amqp://")
    serv.startService()
    client = yield serv.getFactory().whenConnected()

    queues = [{"queue": str(uuid.uuid4()), "passive": True}]
    try:
        yield client.declare_queues(queues)
    except exceptions.BadDeclaration as e:
        assert "queue" == e.obj_type
        assert queues[0] == e.description
        assert isinstance(e.reason, pika.exceptions.ChannelClosed)
    serv.stopService()


@pytest_twisted.inlineCallbacks
def test_declare_exchange_failures():
    """Assert that if an exchange can't be declared, it results in an exception."""
    serv = service.FedoraMessagingService(amqp_url="amqp://")
    serv.startService()
    client = yield serv.getFactory().whenConnected()

    exchanges = [{"exchange": str(uuid.uuid4()), "passive": True}]
    try:
        yield client.declare_exchanges(exchanges)
    except exceptions.BadDeclaration as e:
        assert "exchange" == e.obj_type
        assert exchanges[0] == e.description
        assert isinstance(e.reason, pika.exceptions.ChannelClosed)
    serv.stopService()


@pytest_twisted.inlineCallbacks
def test_declare_binding_failure():
    """Assert that if a binding can't be declared, it results in an exception."""
    serv = service.FedoraMessagingService(amqp_url="amqp://")
    serv.startService()
    client = yield serv.getFactory().whenConnected()

    binding = [
        {"exchange": str(uuid.uuid4()), "queue": str(uuid.uuid4()), "routing_key": "#"}
    ]
    try:
        yield client.bind_queues(binding)
    except exceptions.BadDeclaration as e:
        assert "binding" == e.obj_type
        assert binding[0] == e.description
        assert isinstance(e.reason, pika.exceptions.ChannelClosed)
    serv.stopService()



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for the testing utilities."""
try:
    from unittest import mock
except ImportError:
    import mock
import unittest

from fedora_messaging import testing, api, message


class CustomMessage(api.Message):
    pass


class MockSendsTests(unittest.TestCase):
    """Tests for the :func:`fedora_messaging.testing.mock_sends` function."""

    def test_class(self):
        """Assert all goes well if the message published matches the asserted class."""

        def pub():
            api.publish(api.Message())

        with testing.mock_sends(api.Message):
            pub()

    def test_instance(self):
        """Assert all goes well if the message published matches the asserted instance."""

        def pub():
            api.publish(api.Message())

        with testing.mock_sends(api.Message()):
            pub()

    def test_expected_none(self):
        """Assert all goes well if the message published matches the asserted instance."""

        def pub():
            api.publish(api.Message())

        with self.assertRaises(AssertionError) as cm:
            with testing.mock_sends():
                pub()
        self.assertEqual(
            "Expected 0 messages to be sent, but 1 were sent", cm.exception.args[0]
        )

    def test_mix_class_instance(self):
        """Assert a mix of class and instance works."""

        def pub():
            api.publish(api.Message())
            api.publish(CustomMessage())

        with mock.patch.dict(message._class_to_schema_name, {CustomMessage: "custom"}):
            with testing.mock_sends(api.Message(), CustomMessage):
                pub()

    def test_mix_class_instance_order_matters(self):
        """Assert the order of messages matters."""
        expected_err = (
            "Expected message of type <class 'fedora_messaging.tests.unit.test_testing"
            ".CustomMessage'>, but <class 'fedora_messaging.message.Message'> was sent"
        )

        def pub():
            api.publish(api.Message())
            api.publish(CustomMessage())

        with mock.patch.dict(message._class_to_schema_name, {CustomMessage: "custom"}):
            with self.assertRaises(AssertionError) as cm:
                with testing.mock_sends(CustomMessage, api.Message()):
                    pub()
        self.assertEqual(expected_err, cm.exception.args[0])

    def test_too_many(self):
        """Assert publishing more messages than expected fails with an AssertionError."""

        def pub():
            api.publish(api.Message())
            api.publish(api.Message())

        with self.assertRaises(AssertionError) as cm:
            with testing.mock_sends(api.Message):
                pub()
        self.assertEqual(
            "Expected 1 messages to be sent, but 2 were sent", cm.exception.args[0]
        )

    def test_wrong_type(self):
        """Assert sending the wrong type of message raises an AssertionError."""
        expected_err = (
            "Expected message of type <class 'fedora_messaging.tests.unit.test_testing"
            ".CustomMessage'>, but <class 'fedora_messaging.message.Message'> was sent"
        )

        def pub():
            api.publish(api.Message())

        with self.assertRaises(AssertionError) as cm:
            with testing.mock_sends(CustomMessage):
                pub()
        self.assertEqual(expected_err, cm.exception.args[0])



# coding: utf-8

# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for the :module:`fedora_messaging.api` module."""

import unittest
import mock

from fedora_messaging import api, config
from fedora_messaging.exceptions import PublishException
from fedora_messaging.signals import (
    pre_publish_signal,
    publish_signal,
    publish_failed_signal,
)


class CheckCallbackTests(unittest.TestCase):
    """Tests for :func:`api._check_callback`"""

    def test_method(self):
        """Assert methods are valid."""

        class Callback(object):
            def callback(self):
                return

        class_instance = Callback()

        callback = api._check_callback(class_instance.callback)

        self.assertIs(callback, callback)

    def test_func(self):
        """Assert functions are valid."""

        def callback(message):
            return

        cb = api._check_callback(callback)

        self.assertIs(cb, callback)

    def test_class(self):
        """Assert classes are instantiated."""

        class Callback(object):
            def __call__(self, message):
                return "It worked"

        callback = api._check_callback(Callback)

        self.assertEqual(callback(None), "It worked")

    def test_class_no_call(self):
        """Assert classes are instantiated."""

        class Callback(object):
            pass

        try:
            api._check_callback(Callback)
            self.fail("_check_callback failed to raise an exception")
        except ValueError:
            pass

    def test_class_init_args(self):
        """Assert classes are instantiated."""

        class Callback(object):
            def __init__(self, args):
                self.args = args

        try:
            api._check_callback(Callback)
            self.fail("_check_callback failed to raise a TypeError")
        except TypeError:
            pass

    def test_not_callable(self):
        """Assert a ValueError is raised for things that can't be called."""
        try:
            api._check_callback("")
            self.fail("_check_callback failed to raise an ValueError")
        except ValueError:
            pass


class TwistedConsumeTests(unittest.TestCase):
    """Tests for :func:`api.twisted_consume`"""

    @mock.patch("fedora_messaging.api._twisted_service")
    def test_wrap_bindings(self, mock_service):
        """Assert bindings are always passed to the factory as a list."""

        def callback(msg):
            pass

        api.twisted_consume(callback, {}, {})

        mock_service._service.factory.consume.called_once_with(callback, [{}], {})


@mock.patch("fedora_messaging._session.ConsumerSession")
class ConsumeTests(unittest.TestCase):
    def test_defaults(self, mock_session):
        """Assert that bindings and queues come from the config if not provided."""
        api.consume("test_callback")

        mock_session.return_value.consume.assert_called_once_with(
            "test_callback",
            bindings=config.conf["bindings"],
            queues=config.conf["queues"],
        )

    def test_bindings_dict(self, mock_session):
        """Assert consume wraps bindings in a list of just a plain dict"""
        bindings = {"queue": "q1", "exchange": "e1", "routing_keys": ["#"]}

        api.consume("test_callback", bindings)

        mock_session.assert_called_once()
        mock_session.return_value.consume.assert_called_once_with(
            "test_callback", bindings=[bindings], queues=config.conf["queues"]
        )

    def test_bindings_invalid_type(self, mock_session):
        """Assert bindings are validated and result in a value error if they are invalid."""
        self.assertRaises(ValueError, api.consume, "test_callback", "test_bindings")

    def test_bindings_list_of_dict(self, mock_session):
        """Assert consume is working(bindings type is dict)"""
        bindings = [{"queue": "q1", "exchange": "e1", "routing_keys": ["#"]}]
        mock_session.return_value = mock_session

        api.consume("test_callback", bindings)

        mock_session.assert_called_once()
        mock_session.consume.assert_called_once_with(
            "test_callback", bindings=bindings, queues=config.conf["queues"]
        )

    def test_queues_invalid_type(self, mock_session):
        """Assert bindings are validated and result in a value error if they are invalid."""
        self.assertRaises(ValueError, api.consume, None, None, "should be a dict")

    def test_with_queues(self, mock_session):
        """Assert queues is used over the config if provided."""
        queues = {
            "q1": {
                "durable": True,
                "auto_delete": False,
                "exclusive": False,
                "arguments": {},
            }
        }

        api.consume("test_callback", bindings=[], queues=queues)
        mock_session.return_value.consume.assert_called_once_with(
            "test_callback", bindings=[], queues=queues
        )


class PublishTests(unittest.TestCase):
    def setUp(self):
        self.pre_publish_signal_data = {"called": False, "sender": None, "args": None}
        self.publish_signal_data = {"called": False, "sender": None, "args": None}
        self.publish_failed_signal_data = {
            "called": False,
            "sender": None,
            "args": None,
        }

        @pre_publish_signal.connect
        def pre_publish_signal_handler(sender, **kwargs):
            self.pre_publish_signal_data = {
                "called": True,
                "sender": sender,
                "args": kwargs,
            }

        @publish_signal.connect
        def publish_signal_handler(sender, **kwargs):
            self.publish_signal_data = {
                "called": True,
                "sender": sender,
                "args": kwargs,
            }

        @publish_failed_signal.connect
        def publish_failed_signal_handler(sender, **kwargs):
            self.publish_failed_signal_data = {
                "called": True,
                "sender": sender,
                "args": kwargs,
            }

        self.pre_publish_signal_handler = pre_publish_signal_handler
        self.publish_signal_handler = publish_signal_handler
        self.publish_failed_signal_handler = publish_failed_signal_handler

        self.session_cache_patch = mock.patch("fedora_messaging.api._session_cache")
        self.publisher_session_patch = mock.patch(
            "fedora_messaging._session.PublisherSession"
        )
        self.mock_session_cache = self.session_cache_patch.start()
        self.mock_publisher_session = self.publisher_session_patch.start()
        self.mock_publisher_session.return_value = self.mock_publisher_session
        del self.mock_session_cache.session

    def tearDown(self):
        pre_publish_signal.disconnect(self.pre_publish_signal_handler)
        publish_signal.disconnect(self.publish_signal_handler)
        publish_failed_signal.disconnect(self.publish_failed_signal_handler)
        mock.patch.stopall()

    def test_publish_to_exchange(self):
        """Assert a message can be published to the exchange."""
        message = "test_message"
        exchange = "test_exchange"
        expected_pre_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_failed_signal_data = {
            "called": False,
            "sender": None,
            "args": None,
        }
        api.publish(message, exchange)
        self.mock_publisher_session.assert_called_once_with()
        self.mock_publisher_session.publish.assert_called_once_with(
            message, exchange=exchange
        )
        self.assertEqual(self.pre_publish_signal_data, expected_pre_publish_signal_data)
        self.assertEqual(self.publish_signal_data, expected_publish_signal_data)
        self.assertEqual(
            self.publish_failed_signal_data, expected_publish_failed_signal_data
        )

    @mock.patch.dict(
        "fedora_messaging.config.conf", {"publish_exchange": "test_publich_exchange"}
    )
    def test_publish_to_config_exchange(self):
        """Assert a message can be published to the exchange form config."""
        message = "test_message"
        expected_pre_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_failed_signal_data = {
            "called": False,
            "sender": None,
            "args": None,
        }
        api.publish(message)
        self.mock_publisher_session.assert_called_once_with()
        self.mock_publisher_session.publish.assert_called_once_with(
            message, exchange="test_publich_exchange"
        )
        self.assertEqual(self.pre_publish_signal_data, expected_pre_publish_signal_data)
        self.assertEqual(self.publish_signal_data, expected_publish_signal_data)
        self.assertEqual(
            self.publish_failed_signal_data, expected_publish_failed_signal_data
        )

    def test_session_cache_has_session(self):
        """Assert a TLS vaiable _session_cache contains 'session'."""
        message = "test_message"
        exchange = "test_exchange"
        expected_pre_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_failed_signal_data = {
            "called": False,
            "sender": None,
            "args": None,
        }
        self.mock_session_cache.session = self.mock_publisher_session
        api.publish(message, exchange)
        self.mock_publisher_session.assert_not_called()
        self.mock_publisher_session.publish.assert_called_once_with(
            message, exchange=exchange
        )
        self.assertEqual(self.pre_publish_signal_data, expected_pre_publish_signal_data)
        self.assertEqual(self.publish_signal_data, expected_publish_signal_data)
        self.assertEqual(
            self.publish_failed_signal_data, expected_publish_failed_signal_data
        )

    def test_publish_failed(self):
        """Assert an exception is raised when message can't be published."""
        message = "test_message"
        exchange = "test_exchange"
        expected_exception = PublishException(reason="Unable to publish message")
        expected_pre_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_signal_data = {"called": False, "sender": None, "args": None}
        expected_publish_failed_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message, "reason": expected_exception},
        }
        self.mock_publisher_session.publish.side_effect = expected_exception
        self.assertRaises(
            type(expected_exception), api.publish, message, exchange=exchange
        )
        self.mock_publisher_session.assert_called_once_with()
        self.mock_publisher_session.publish.assert_called_once_with(
            message, exchange=exchange
        )
        self.assertEqual(self.pre_publish_signal_data, expected_pre_publish_signal_data)
        self.assertEqual(self.publish_signal_data, expected_publish_signal_data)
        self.assertEqual(
            self.publish_failed_signal_data, expected_publish_failed_signal_data
        )
